/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.fortress.solver;

import ru.ispras.fortress.solver.engine.smt.BoolectorSolver;
import ru.ispras.fortress.solver.engine.sat.Sat4jSolver;
import ru.ispras.fortress.solver.engine.smt.Cvc4Solver;
import ru.ispras.fortress.solver.engine.smt.Z3Solver;

/**
 * The {@link SolverId} enumeration specifies which solver should be used to solve
 * a particular constraint.
 *
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public enum SolverId {
  /**
   * Z3 solver by Microsoft Research.
   *
   * <p>It processes a text file with SMT-LIB code and prints results to the output stream.</p>
   */
  Z3_TEXT {
    protected Solver newSolver() {
      return new Z3Solver();
    }
  },

  /**
   * CVC4 solver, a joint project of Stanford University and University of Iowa.
   *
   * <p>It processes a text file with SMT-LIB code and prints results to the output stream.</p>
   */
  CVC4_TEXT {
    protected Solver newSolver() {
      return new Cvc4Solver();
    }
  },

  BOOLECTOR_TEXT {
    protected Solver newSolver() {
      return new BoolectorSolver();
    }
  },
  /**
   * Sat4j solver, a Java implementation of MiniSAT (extended w/ other engines).
   */
  SAT4J_API {
    protected Solver newSolver() {
      return new Sat4jSolver();
    }
  };

  private Solver solver = null;

  /**
   * Returns a singleton instance of a corresponding solver.
   *
   * @return Solver instance.
   */
  public final Solver getSolver() {
    if (null == solver) {
      solver = newSolver();
    }

    return solver;
  }

  protected abstract Solver newSolver();
}
