/*
 * Copyright 2017-2020 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.fortress.solver.constraint;

import static ru.ispras.fortress.logic.BitBlaster.AND;
import static ru.ispras.fortress.logic.BitBlaster.BVADD;
import static ru.ispras.fortress.logic.BitBlaster.BVAND;
import static ru.ispras.fortress.logic.BitBlaster.BVASHL;
import static ru.ispras.fortress.logic.BitBlaster.BVASHR;
import static ru.ispras.fortress.logic.BitBlaster.BVCONCAT;
import static ru.ispras.fortress.logic.BitBlaster.BVLSHL;
import static ru.ispras.fortress.logic.BitBlaster.BVLSHR;
import static ru.ispras.fortress.logic.BitBlaster.BVNAND;
import static ru.ispras.fortress.logic.BitBlaster.BVNEG;
import static ru.ispras.fortress.logic.BitBlaster.BVNOR;
import static ru.ispras.fortress.logic.BitBlaster.BVOR;
import static ru.ispras.fortress.logic.BitBlaster.BVROL;
import static ru.ispras.fortress.logic.BitBlaster.BVROR;
import static ru.ispras.fortress.logic.BitBlaster.BVSGE;
import static ru.ispras.fortress.logic.BitBlaster.BVSGT;
import static ru.ispras.fortress.logic.BitBlaster.BVSLE;
import static ru.ispras.fortress.logic.BitBlaster.BVSLT;
import static ru.ispras.fortress.logic.BitBlaster.BVUGE;
import static ru.ispras.fortress.logic.BitBlaster.BVUGT;
import static ru.ispras.fortress.logic.BitBlaster.BVULE;
import static ru.ispras.fortress.logic.BitBlaster.BVULT;
import static ru.ispras.fortress.logic.BitBlaster.BVXOR;
import static ru.ispras.fortress.logic.BitBlaster.EQ;
import static ru.ispras.fortress.logic.BitBlaster.EQ_CONSTANT;
import static ru.ispras.fortress.logic.BitBlaster.FALSE;
import static ru.ispras.fortress.logic.BitBlaster.IMPL;
import static ru.ispras.fortress.logic.BitBlaster.ITE;
import static ru.ispras.fortress.logic.BitBlaster.NOTEQ;
import static ru.ispras.fortress.logic.BitBlaster.NOTEQ_CONSTANT;
import static ru.ispras.fortress.logic.BitBlaster.OR;
import static ru.ispras.fortress.logic.BitBlaster.XOR;

import java.math.BigInteger;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.function.IntSupplier;
import java.util.stream.IntStream;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.logic.BitBlaster.Operand;
import ru.ispras.fortress.solver.engine.sat.Initializer;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * {@link Sat4jFormulaEncoder} implements a SAT4J constraint encoder.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class Sat4jFormulaEncoder implements ConstraintEncoder {
  /** Using directly ISolver instead of Sat4jFormula.Builder slows down performance. */
  private final Sat4jFormula.Builder builder;

  /** Maps variables to their lower bit indices. */
  private final Map<Variable, Integer> indices;
  /** Maps variables to their bit-usage masks. */
  private final Map<Variable, BitVector> masks;

  /** Index of a boolean variable to be allocated next. */
  private int index;

  /** Returns the index of a new boolean variable. */
  private final IntSupplier newIndex = new IntSupplier() {
    @Override
    public int getAsInt() {
      return index++;
    }
  };

  /** Used to fill unused variable fields. */
  private final Initializer initializer;

  public Sat4jFormulaEncoder(final Initializer initializer) {
    InvariantChecks.checkNotNull(initializer);

    this.indices = new LinkedHashMap<>();
    this.masks = new LinkedHashMap<>();

    this.builder = new Sat4jFormula.Builder(indices, masks);

    // In SAT4J, a boolean variable index should be positive:
    // x[i] and ~x[i] are mapped to +i and -i respectively.
    this.index = 1;
    this.initializer = initializer;
  }

  public Sat4jFormulaEncoder() {
    this(Initializer.RANDOM);
  }

  public Sat4jFormulaEncoder(final Sat4jFormulaEncoder other) {
    InvariantChecks.checkNotNull(other);

    this.builder = new Sat4jFormula.Builder(other.builder);
    this.indices = new LinkedHashMap<>(other.indices);
    this.masks = new LinkedHashMap<>(other.masks);
    this.index = other.index;
    this.initializer = other.initializer;
  }

  @Override
  public void addNode(final Node node) {
    encodeConstants(node);
    encode(node, 0, false);
  }

  @Override
  public Constraint encode() {
    final ConstraintBuilder constraintBuilder = new ConstraintBuilder();

    constraintBuilder.setKind(ConstraintKind.SAT4J_BASED);
    constraintBuilder.addVariables(indices.keySet());
    constraintBuilder.setInnerRep(builder.build());

    return constraintBuilder.build();
  }

  @Override
  public Sat4jFormulaEncoder clone() {
    return new Sat4jFormulaEncoder(this);
  }

  private void encodeConstants(final Node node) {
    for (final NodeVariable lhs : ExprUtils.getVariables(node)) {
      final int preIndex = index;
      final int varIndex = getVariableIndex(lhs);

      // If the variable has not been encoded yet.
      if (varIndex >= preIndex && lhs.getVariable().hasValue()) {
        setUsedBits(lhs.getVariable());

        final BigInteger value = Sat4jUtils.getInteger(lhs.getVariable().getData());
        final NodeValue rhs = NodeValue.newInteger(value);

        builder.addAll(EQ_CONSTANT.encode(encodeOperands(lhs, rhs), newIndex));
      }
    }
  }

  private void encode(final Node node, int flag, final boolean negation) {
    if (!ExprUtils.isCondition(node)) {
      encodeTermAsBool(node, flag, negation);
    } else if (ExprUtils.isAtomicCondition(node)) {
      encodePredicate((NodeOperation) node, flag, negation);
    } else {
      encodeConnective((NodeOperation) node, flag, negation);
    }
  }

  private void encodeTermAsBool(final Node node, final int flag, final boolean negation) {
    final Operand operand = encodeTerm(node);

    if (operand.isValue()) {
      final Operand operands[] = new Operand[] {};
      FALSE.encode(operands, flag, newIndex, (operand.value.intValue() != 0) ^ negation);
    } else {
      final Operand operands[] = new Operand[] { operand, new Operand(BigInteger.valueOf(0)) };
      NOTEQ_CONSTANT.encode(operands, flag, newIndex, negation);
    }
  }

  private void encodePredicate(final NodeOperation node, final int flag, final boolean negation) {
    final Operand[] operands = encodeOperands(node.getOperands());

    if (ExprUtils.isOperation(node, StandardOperation.EQ)) {
      builder.addAll(EQ.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.NOTEQ)) {
      builder.addAll(NOTEQ.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.BVULE, StandardOperation.LESSEQ)) {
      builder.addAll(BVULE.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.BVULT, StandardOperation.LESS)) {
      builder.addAll(BVULT.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.BVUGE, StandardOperation.GREATEREQ)) {
      builder.addAll(BVUGE.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.BVUGT, StandardOperation.GREATER)) {
      builder.addAll(BVUGT.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.BVSLE)) {
      builder.addAll(BVSLE.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.BVSLT)) {
      builder.addAll(BVSLT.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.BVSGE)) {
      builder.addAll(BVSGE.encode(operands, flag, newIndex, negation));
    } else if (ExprUtils.isOperation(node, StandardOperation.BVSGT)) {
      builder.addAll(BVSGT.encode(operands, flag, newIndex, negation));
    } else {
      reportUnknownNode(node);
    }
  }

  private void encodeConnective(final NodeOperation node, final int flag, final boolean negation) {
    if (ExprUtils.isOperation(node, StandardOperation.NOT)) {
      encode(node.getOperand(0), flag, !negation);
    } else if (ExprUtils.isOperation(node, StandardOperation.AND)) {
      encodeConjunction(node, flag, negation, false);
    } else if (ExprUtils.isOperation(node, StandardOperation.OR)) {
      encodeDisjunction(node, flag, negation, false);
    } else if (ExprUtils.isOperation(node, StandardOperation.XOR)) {
      encodeExclusiveDisjunction(node, flag, negation);
    } else if (ExprUtils.isOperation(node, StandardOperation.IMPL)) {
      encodeImplication(node, flag, negation);
    } else if (ExprUtils.isOperation(node, StandardOperation.ITE)) {
      encodeIfThenElse(node, flag, negation);
    } else {
      reportUnknownNode(node);
    }
  }

  private Operand encodeTerm(final Node node) {
    InvariantChecks.checkFalse(ExprUtils.isCondition(node));

    // Node is a constant.
    if (ExprUtils.isValue(node)) {
      return new Operand(Sat4jUtils.getInteger(node));
    }

    // Node is a variable or an extract, i.e. x[h:l].
    if (Sat4jUtils.getVariable(node) != null) {
      final int index = getVariableIndex(node);
      final int size  = Sat4jUtils.getBitSize(node);
      final int lower = Sat4jUtils.getLowerBit(node);

      setUsedBits(node);

      return new Operand(index + lower, true, size);
    }

    // Node is an operation, i.e. x + y.
    final NodeOperation operation = (NodeOperation) node;
    final Operand[] operands = encodeOperands(1, operation.getOperands());

    // Introduce a new variable, i.e. u = x + y.
    final int size  = Sat4jUtils.getBitSize(node);
    final int index = getVariableIndex(size);
    operands[0] = new Operand(index, true, size);

    if (ExprUtils.isOperation(node, StandardOperation.BVNEG, StandardOperation.BVNOT)) {
      BVNEG.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVCONCAT)) {
      BVCONCAT.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVAND)) {
      BVAND.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVOR)) {
      BVOR.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVXOR)) {
      BVXOR.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVNAND)) {
      BVNAND.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVNOR)) {
      BVNOR.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVLSHL)) {
      BVLSHL.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVLSHR)) {
      BVLSHR.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVASHL)) {
      BVASHL.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVASHR)) {
      BVASHR.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVROL)) {
      BVROL.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVROR)) {
      BVROR.encode(operands, newIndex);
    } else if (ExprUtils.isOperation(node, StandardOperation.BVADD, StandardOperation.ADD)) {
      BVADD.encode(operands, newIndex);
    } else {
      reportUnknownNode(node);
    }
    return null;
  }

  private void encodeConjunction(
      final NodeOperation node,
      final int flag,
      final boolean externalNegation,
      final boolean internalNegation) {
    // De Morgan's law: ~(x & y) == (~x | ~y).
    if (externalNegation) {
      encodeDisjunction(node, flag, false, true);
      return;
    }

    int flagIndex = flag != 0 ? index : 0;
    int flagDelta = flag != 0 ? 1 : 0;

    // f <=> (f[0] & ... & f[n-1]).
    if (flag != 0) {
      builder.addAll(AND.encode(newBoolOperands(node.getOperandCount()), flag, newIndex));
    }

    // f[i] <=> encode(operand[i])
    for (final Node operand : node.getOperands()) {
      // Another option: operand, -flagIndex, !internalNegation.
      encode(operand, flagIndex, internalNegation);
      flagIndex += flagDelta;
    }
  }

  private void encodeDisjunction(
      final NodeOperation node,
      final int flag,
      final boolean externalNegation,
      final boolean internalNegation) {
    // De Morgan's law: ~(x | y) == (~x & ~y).
    if (externalNegation) {
      encodeConjunction(node, flag, false, true);
      return;
    }

    int flagIndex = index;

    // f <=> (f[0] | ... | f[n-1]).
    builder.addAll(OR.encode(newBoolOperands(node.getOperandCount()), flag, newIndex));

    // f[i] <=> encode(operand[i])
    for (final Node operand : node.getOperands()) {
      encode(operand, flagIndex, internalNegation);
      flagIndex++;
    }
  }

  private void encodeExclusiveDisjunction(
      final NodeOperation node,
      final int flag,
      final boolean negation) {
    int flagIndex = index;

    // f <=> (f[0] ^ ... ^ f[n-1]).
    builder.addAll(XOR.encode(newBoolOperands(node.getOperandCount()), flag, newIndex, negation));

    // f[i] <=> encode(operand[i])
    for (final Node operand : node.getOperands()) {
      encode(operand, flagIndex, false);
      flagIndex++;
    }
  }

  private void encodeImplication(
      final NodeOperation node,
      final int flag,
      final boolean negation) {
    int flagIndex = index;

    // f <=> (f[0] -> f[1]).
    InvariantChecks.checkTrue(node.getOperandCount() == 2);
    builder.addAll(IMPL.encode(newBoolOperands(2), flag, newIndex, negation));

    // f[i] <=> encode(operand[i])
    for (final Node operand : node.getOperands()) {
      encode(operand, flagIndex, false);
      flagIndex++;
    }
  }

  private void encodeIfThenElse(
      final NodeOperation node,
      final int flag,
      final boolean negation) {
    int flagIndex = index;

    // f <=> (f[0] ? f[1] : f[2]).
    InvariantChecks.checkTrue(node.getOperandCount() == 3);
    builder.addAll(ITE.encode(newBoolOperands(3), flag, newIndex, negation));

    // f[i] <=> encode(operand[i]).
    for (final Node operand : node.getOperands()) {
      encode(operand, flagIndex, false);
      flagIndex++;
    }
  }

  private Operand[] encodeOperands(final int extra, final List<Node> nodes) {
    final Operand[] operands = new Operand[nodes.size() + extra];
    IntStream.range(0, nodes.size()).forEach(i -> operands[i + extra] = encodeTerm(nodes.get(i)));
    return operands;
  }

  private Operand[] encodeOperands(final List<Node> nodes) {
    return encodeOperands(0, nodes);
  }

  private Operand[] encodeOperands(final Node... nodes) {
    return encodeOperands(0, Arrays.asList(nodes));
  }

  private Operand[] newBoolOperands(final int count) {
    return IntStream.range(0, count).mapToObj(
        i -> new Operand(newIndex.getAsInt(), 1)).toArray(Operand[]::new);
  }

  private int getVariableIndex(final Node node) {
    final Variable variable = Sat4jUtils.getVariable(node);
    InvariantChecks.checkNotNull(variable);

    final Integer oldIndex = indices.get(variable);

    if (oldIndex != null) {
      return oldIndex;
    }

    final int newIndex = getVariableIndex(variable.getType().getSize());
    indices.put(variable, newIndex);

    return newIndex;
  }

  private int getVariableIndex(final int size) {
    final int newIndex = index;
    index += size;
    return newIndex;
  }

  private BitVector getVariableMask(final Variable variable) {
    BitVector mask = masks.get(variable);

    if (mask == null) {
      masks.put(variable, mask = BitVector.newEmpty(Sat4jUtils.getBitSize(variable)));
    }

    return mask;
  }

  private void setUsedBits(final Node node) {
    final Variable variable = Sat4jUtils.getVariable(node);
    if (variable == null) {
      return;
    }

    final BitVector mask = getVariableMask(variable);
    for (int i = Sat4jUtils.getLowerBit(node); i <= Sat4jUtils.getUpperBit(node); i++) {
      mask.setBit(i, true);
    }
  }

  private void setUsedBits(final Variable variable) {
    final BitVector mask = getVariableMask(variable);
    mask.setAll();
  }

  private void reportUnknownNode(final Node node) {
    InvariantChecks.checkTrue(false,
        String.format("Encoding failed: unknown node %s", node));
  }
}
