/*
 * Copyright 2014-2020 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.fortress.solver.constraint;

import java.math.BigInteger;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

public final class Sat4jUtils {
  private Sat4jUtils() {}

  public static BigInteger getInteger(final Data data) {
    switch (data.getType().getTypeId()) {
      case BIT_VECTOR:
        return data.getBitVector().bigIntegerValue(false);

      case LOGIC_INTEGER:
        return data.getInteger();

      case LOGIC_BOOLEAN:
        return data.getBoolean() ? BigInteger.ONE : BigInteger.ZERO;

      default:
        InvariantChecks.checkTrue(false);
        return null;
    }
  }

  public static BigInteger getInteger(final Node expr) {
    checkConstantValue(expr);
    final NodeValue value = (NodeValue) expr;

    switch (value.getDataTypeId()) {
      case BIT_VECTOR:
        return value.getBitVector().bigIntegerValue(false);

      case LOGIC_INTEGER:
        return value.getInteger();

      case LOGIC_BOOLEAN:
        return value.getBoolean() ? BigInteger.ONE : BigInteger.ZERO;

      default:
        InvariantChecks.checkTrue(false);
        return null;
    }
  }

  private static void checkConstantValue(final Node expr) {
    if (!ExprUtils.isValue(expr)) {
      throw new IllegalStateException(String.format("%s is not a constant value.", expr));
    }
  }

  public static Variable getVariable(final Node node) {
    switch (node.getKind()) {
      case VARIABLE:
        final NodeVariable variable = (NodeVariable) node;
        return variable.getVariable();

      case OPERATION:
        InvariantChecks.checkTrue(ExprUtils.isOperation(node, StandardOperation.BVEXTRACT));
        final NodeOperation operation = (NodeOperation) node;
        return getVariable(operation.getOperand(2));

      default:
        return null;
    }
  }

  public static int getLowerBit(final Node node) {
    switch (node.getKind()) {
      case VALUE:
      case VARIABLE:
        return 0;

      case OPERATION:
        InvariantChecks.checkTrue(ExprUtils.isOperation(node, StandardOperation.BVEXTRACT));
        final NodeOperation operation = (NodeOperation) node;
        final NodeValue lowerBitValue = (NodeValue) operation.getOperand(1);
        return lowerBitValue.getInteger().intValue();

      default:
        InvariantChecks.checkTrue(false);
        return -1;
    }
  }

  public static int getUpperBit(final Node node) {
    switch (node.getKind()) {
      case VALUE:
        final NodeValue value = (NodeValue) node;
        final int valueBitSize = value.getDataType().getSize();
        return valueBitSize > 0 ? valueBitSize - 1 : Integer.SIZE - 1;

      case VARIABLE:
        final NodeVariable variable = (NodeVariable) node;
        return getBitSize(variable.getVariable()) - 1;

      case OPERATION:
        InvariantChecks.checkTrue(ExprUtils.isOperation(node, StandardOperation.BVEXTRACT));
        final NodeOperation operation = (NodeOperation) node;
        final NodeValue upperBitValue = (NodeValue) operation.getOperand(0);
        return upperBitValue.getInteger().intValue();

      default:
        InvariantChecks.checkTrue(false);
        return -1;
    }
  }

  public static int getBitSize(final Variable variable) {
    final int bitSize = variable.getType().getSize();
    return bitSize > 0 ? bitSize : Integer.SIZE;
  }

  public static int getBitSize(final Node node) {
    return (getUpperBit(node) - getLowerBit(node)) + 1;
  }
}
