/*
 * Copyright 2017-2020 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.fortress.solver.engine.sat;

import java.util.Map;
import org.sat4j.minisat.SolverFactory;
import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.IProblem;
import org.sat4j.specs.ISolver;
import org.sat4j.specs.IVecInt;
import org.sat4j.specs.TimeoutException;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.solver.Solver;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.SolverResultBuilder;
import ru.ispras.fortress.solver.constraint.Constraint;
import ru.ispras.fortress.solver.constraint.ConstraintKind;
import ru.ispras.fortress.solver.constraint.Sat4jFormula;
import ru.ispras.fortress.solver.function.Function;
import ru.ispras.fortress.solver.function.FunctionTemplate;

/**
 * {@link Sat4jSolver} is a SAT-based {@code QF_BV} constraint solver.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class Sat4jSolver implements Solver {
  private static final String NAME = "Sat4j";
  private static final String DESCRIPTION = "Solves constraints using the Sat4j solver";

  private Initializer initializer = Initializer.RANDOM;

  public void setInitializer(final Initializer initializer) {
    this.initializer = initializer;
  }

  @Override
  public String getName() {
    return NAME;
  }

  @Override
  public String getDescription() {
    return DESCRIPTION;
  }

  @Override
  public boolean isSupported(final ConstraintKind kind) {
    return kind == ConstraintKind.SAT4J_BASED;
  }

  @Override
  public boolean isGeneric() {
    return false;
  }

  @Override
  public SolverResult solve(final Constraint constraint) {
    final Sat4jFormula formula = (Sat4jFormula) constraint.getInnerRep();
    final ISolver solver = SolverFactory.newDefault();
    final SolverResultBuilder resultBuilder = new SolverResultBuilder(Status.UNKNOWN);

    // Construct the problem for Sat4j.
    try {
      for (final IVecInt clause : formula.getClauses()) {
        solver.addClause(clause);
      }
    } catch (final ContradictionException e) {
      resultBuilder.setStatus(Status.UNSAT);
      resultBuilder.addError(e.getMessage());
      return resultBuilder.build();
    }

    // Solve the problem.
    try {
      if (!solver.isSatisfiable()) {
        return resultBuilder.build();
      }
    } catch (final TimeoutException e) {
      resultBuilder.addError(e.getMessage());
      return resultBuilder.build();
    }

// FIXME: What if a model is not required.
//    if (mode == Solver.Mode.SAT) {
//      return SolverResult.newSat();
//    }

    // Return the solution.
    return decode(formula, solver);
  }

  private SolverResult decode(final Sat4jFormula formula, final IProblem solution) {
    final SolverResultBuilder resultBuilder = new SolverResultBuilder(SolverResult.Status.SAT);

    for (final Map.Entry<Variable, Integer> entry : formula.getIndices().entrySet()) {
      final Variable variable = entry.getKey();
      final int index = entry.getValue();
      final BitVector mask = formula.getMasks().get(variable);

      // Initialize the variable (e.g. with a random value).
      final BitVector value = initializer.getValue(variable.getType().getSize());

      // Reassign the bits used in the constraint.
      for (int i = 0; i < variable.getType().getSize(); i++) {
        if (mask.getBit(i)) {
          value.setBit(i, solution.model(index + i));
        }
      }

      final String name = variable.getName();
      final Data data = Data.newBitVector(value);

      resultBuilder.addVariable(new Variable(name, data));
    }

    return resultBuilder.build();
  }

  @Override
  public boolean addCustomOperation(final Function function) {
    throw new UnsupportedOperationException();
  }

  @Override
  public boolean addCustomOperation(final FunctionTemplate template) {
    throw new UnsupportedOperationException();
  }

  @Override
  public String getSolverPath() {
    throw new UnsupportedOperationException();
  }

  @Override
  public void setSolverPath(final String value) {
    throw new UnsupportedOperationException();
  }
}
