/*
 * Copyright 2020 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.fortress.solver.engine.smt;

import ru.ispras.fortress.expression.StandardOperation;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Collections;
import java.util.List;

/**
 * The BoolectorSolver class implements logic of a constraint solver that uses the Boolector tool.
 * The constraint is translated to STM-LIB code that is then saved to a file and processed
 * to the tool.
 *
 * @author <a href="mailto:youda98@gmail.com">Maxim Zimnyukov</a>
 */

public final class BoolectorSolver extends SmtTextSolver {
  private static final String NAME = "Boolector (text-based interface)";

  private static final String DESCRIPTION =
      "Solves constraints using the Boolector solver. "
          + "Interacts with the solver via text files and command line.";

  private static final String ENV_VAR_NAME = "BOOLECTOR_PATH";

  private static final String HEADER = "(set-logic QF_AUFBV)";

  /**
   * Construct a BoolectorSolver object.
   */
  public BoolectorSolver() {
    super(NAME, DESCRIPTION, ENV_VAR_NAME);
    initBoolectorOperations();
  }

  @Override
  protected List<String> getHeader() {
    return Collections.singletonList(HEADER);
  }

  @Override
  public Reader invokeSolver(final String path) throws IOException {
    final Process process =
        new ProcessBuilder(getSolverPath(), "-m", path).start();

    return new BufferedReader(new InputStreamReader(process.getInputStream()));
  }

  private void initBoolectorOperations() {
    addStandardOperation(StandardOperation.REM, "rem");
  }
}
