/*
 * Copyright 2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.fortress.transformer;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class IntToBitVecRuleSet {

  public static NodeTransformer getRuleSet(Integer intBitLength) {

    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule mulToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.MUL);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVMUL, op.getOperands());
      }
    };

    final TransformerRule addToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.ADD);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVADD, op.getOperands());
      }
    };

    final TransformerRule subToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.SUB);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVSUB, op.getOperands());
      }
    };

    final TransformerRule divToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.DIV);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVSDIV, op.getOperands());
      }
    };

    final TransformerRule remToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.REM);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVSREM, op.getOperands());
      }
    };

    final TransformerRule modToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.MOD);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVSMOD, op.getOperands());
      }
    };

    final TransformerRule lessEqToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.LESSEQ);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVSLE, op.getOperands());
      }
    };

    final TransformerRule lessToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.LESS);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVSLT, op.getOperands());
      }
    };

    final TransformerRule greaterToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.GREATER);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVSGT, op.getOperands());
      }
    };

    final TransformerRule greaterEqToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.GREATEREQ);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVSGE, op.getOperands());
      }
    };

    final TransformerRule minusToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.MINUS);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(StandardOperation.BVNEG, op.getOperands());
      }
    };

    final TransformerRule plusToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.PLUS);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return op.getOperands().get(0);
      }
    };

    final TransformerRule operandsToBitVec = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node);
      }

      @Override
      public Node apply(final Node node) {
        final NodeOperation op = (NodeOperation) node;

        return new NodeOperation(op.getOperationId(), castArgsToBitVec(op, intBitLength));
      }
    };

    for (final StandardOperation operation : StandardOperation.values()) {
        if (!StandardOperation.getBvOperandOperations().contains(operation)) {
          transformer.addRule(operation, operandsToBitVec);
        }
    }
    transformer.addRule(StandardOperation.STORE, operandsToBitVec);
    transformer.addRule(StandardOperation.MUL, mulToBitVec);
    transformer.addRule(StandardOperation.ADD, addToBitVec);
    transformer.addRule(StandardOperation.SUB, subToBitVec);
    transformer.addRule(StandardOperation.DIV, divToBitVec);
    transformer.addRule(StandardOperation.REM, remToBitVec);
    transformer.addRule(StandardOperation.MOD, modToBitVec);
    transformer.addRule(StandardOperation.LESSEQ, lessEqToBitVec);
    transformer.addRule(StandardOperation.LESS, lessToBitVec);
    transformer.addRule(StandardOperation.GREATEREQ, greaterEqToBitVec);
    transformer.addRule(StandardOperation.GREATER, greaterToBitVec);
    transformer.addRule(StandardOperation.PLUS, plusToBitVec);
    transformer.addRule(StandardOperation.MINUS, minusToBitVec);

    return transformer;
  }

  private static NodeVariable intVarToBitVec(NodeVariable node, Integer intBitLength) {
    return NodeVariable.newBitVector(node.getName() + "_bv", intBitLength);
  }

  private static NodeValue intValToBitVec(NodeValue node, Integer intBitLength) {
    return NodeValue.newBitVector(node.getInteger(), intBitLength);
  }

  private static List<Node> castArgsToBitVec(NodeOperation node, Integer intBitLength) {
    final List<Node> casted = new LinkedList<>();
    for (final Node operand : node.getOperands()) {

      final boolean isIntOp = operand.isType(DataType.INTEGER);
      final boolean isMapOp = operand.isType(DataTypeId.MAP);

      if (ExprUtils.isValue(operand) && isIntOp) {

        casted.add(intValToBitVec((NodeValue) operand, intBitLength));
      } else if (ExprUtils.isVariable(operand) && isIntOp) {

        casted.add(intVarToBitVec((NodeVariable) operand, intBitLength));
      } else if (ExprUtils.isVariable(operand) &&
          operand.isType(DataTypeId.MAP)) {
        Object[] params = operand.getDataType().getParameters();

        if (params.length != 2) {
          casted.add(operand);
        } else {
          DataType keyType = (DataType) params[0];
          DataType valueType = (DataType) params[1];
          final boolean isIntKey = keyType.isTypeId(DataTypeId.LOGIC_INTEGER);
          final boolean isIntVal = valueType.isTypeId(DataTypeId.LOGIC_INTEGER);
          boolean modified = false;

          if (isIntKey) {

            keyType = DataType.bitVector(intBitLength);
            modified = true;
          }
          if (isIntVal) {

            valueType = DataType.bitVector(intBitLength);
            modified = true;
          }

          if (modified) {
            casted.add(NodeVariable.newMap(
                ((NodeVariable) operand).getName() + "_bv", keyType, valueType));
          } else {
            casted.add(operand);
          }
        }
      } else if (ExprUtils.isValue(operand) && isMapOp) {

          DataMap map = ((NodeValue) operand).getMap();

          final DataType keyType = map.getKeyType();
          final DataType valueType = map.getValueType();
          final boolean isIntKey = keyType.isTypeId(DataTypeId.LOGIC_INTEGER);
          final boolean isIntVal = valueType.isTypeId(DataTypeId.LOGIC_INTEGER);

          if (isIntKey) {

            final DataMap newMap;

            if (isIntVal) {

              newMap =
                  new DataMap(DataType.bitVector(intBitLength), DataType.bitVector(intBitLength));
              for (Map.Entry<Data, Data> entry : newMap.entrySet()) {

              newMap.put(
                  Data.newBitVector(entry.getKey().getInteger(), intBitLength),
                  Data.newBitVector(entry.getValue().getInteger(), intBitLength));
              }
            } else {

              newMap = new DataMap(map.getValueType(), DataType.bitVector(intBitLength));
              for (Map.Entry<Data, Data> entry : newMap.entrySet()) {

                newMap.put(
                    Data.newBitVector(entry.getValue().getInteger(), intBitLength),
                    entry.getValue());
              }
            }
            casted.add(NodeValue.newMap(newMap));
          } else {

            if (isIntVal) {

              final DataMap newMap =
                  new DataMap(map.getValueType(), DataType.bitVector(intBitLength));
              newMap.replaceAll((k, v) -> Data.newBitVector(v.getInteger(), intBitLength));
              casted.add(NodeValue.newMap(newMap));
            } else {
              casted.add(operand);
            }
          }
      } else {

        casted.add(operand);
      }
    }
    return casted;
  }
}
