/*
 * Copyright 2012-2018 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.fortress.data.types.bitvector;

import org.junit.Assert;

/**
 * The {@code TestUtils} class contains utility methods for test-related purposes.
 *
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public final class TestUtils {

  private TestUtils() {
    /* Empty. */
  }

  private static final boolean OUTPUT_DEBUG_STRINGS = true;

  private static final String SEPARATOR = "/"
      + "******************************************************************" + "%n* %-64s*%n"
      + "*******************************************************************" + "/";

  /**
   * If debug output is enabled, prints the specified text to standard output.
   * @param text The text to be printed.
   */
  public static void trace(final String text) {
    if (OUTPUT_DEBUG_STRINGS) {
      System.out.println(text);
    }
  }

  /**
   * Prints the text in a header-formatted form.
   * @param title The text to be printed as a header.
   */
  public static void header(final String title) {
    trace(String.format(SEPARATOR, title));
  }

  /**
   * Checks if the specified bit vector has the specified binary value.
   * @param current The bit vector to be checked.
   * @param expected The binary value string.
   */
  public static void checkBitVector(final BitVector current, final BitVector expected) {
    checkBitVector(current, expected.toBinString());
  }

  /**
   * Checks if the specified bit vector has the specified binary value.
   * @param current The bit vector to be checked.
   * @param expected The binary value string.
   */
  public static void checkBitVector(final BitVector current, final String expected) {
    final String dataString = current.toBinString();
    trace(String.format("Data:     %32s%nExpected: %32s", dataString, expected));

    Assert.assertEquals(
        String.format("Values do not match. %s (data) != %s (expected)", dataString, expected),
        expected, dataString);
  }

  /**
   * Checks if the specified bit vector has the specified binary value.
   * @param current The bit vector to be checked.
   * @param expected The integer value that binary form is used.
   */
  public static void checkBitVector(final BitVector current, final int expected) {
    checkBitVector(current, toBinString(expected, current.getBitSize()));
  }

  /**
   * Checks if the specified bit vector has the specified binary value.
   * @param data The bit vector to be checked.
   * @param expected The long value that binary form is used.
   */
  public static void checkBitVector(final BitVector data, final long expected) {
    checkBitVector(data, TestUtils.toBinString(expected, data.getBitSize()));
  }

  /**
   * Creates a string form with the specified size of integer value.
   * @param value The value to be transformed.
   * @param bitSize The string size.
   * @return String form of the specified value, the form is of the specified size.
   */
  public static String toBinString(final int value, final int bitSize) {
    final String binStr = Integer.toBinaryString(value);

    if (binStr.length() > bitSize) {
      return binStr.substring(binStr.length() - bitSize, binStr.length());
    }

    final int count = bitSize - binStr.length();
    return (count > 0) ? String.format("%0" + count + "d", 0) + binStr : binStr;
  }

  /**
   * Creates a string form with the specified size of integer value.
   * @param value The value to be transformed.
   * @param bitSize The string size.
   * @return String form of the specified value, the form is of the specified size.
   */
  public static String toBinString(final long value, final int bitSize) {
    final String binStr = Long.toBinaryString(value);

    if (binStr.length() > bitSize) {
      return binStr.substring(binStr.length() - bitSize, binStr.length());
    }

    final int count = bitSize - binStr.length();
    return (count > 0) ? String.format("%0" + count + "d", 0) + binStr : binStr;
  }

  /**
   * Returns the comparison result in a string form.
   * @param val The result of the comparison.
   * @return The string form of the comparison result.
   */
  public static String comparisonMsg(final int val) {
    switch (val) {
      case -1:
        return "<";
      case 0:
        return "==";
      case 1:
        return ">";
      default:
    }
    return "failed to compare with";
  }

  /**
   * Checks if comparison result for the specified objects is the same as expected.
   * @param lhs The left hand-side object.
   * @param rhs The right hand-side object.
   * @param expected The expected result of objects comparison.
   */
  public static void checkComparison(final BitVector lhs, final BitVector rhs, final int expected) {
    final String lstr = lhs.toBinString();
    final String rstr = rhs.toBinString();
    trace(String.format("Comparison lhs: %32s%nComparison rhs: %32s", lstr, rstr));

    int actual = lhs.compareTo(rhs);

    Assert.assertEquals(String
        .format("Wrong comparison result. %s (lhs) %s %s (rhs), expected %s", lstr,
            comparisonMsg(actual), rstr, comparisonMsg(expected)), actual, expected);
  }
}
