/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.fortress.solver;

import org.junit.Assert;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.Variable;

import java.util.Iterator;
import java.util.TreeMap;

/**
 * The {@code SolverResultChecker} class contains utility methods to perform operations
 * with results are returned by a SAT solver.
 *
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public final class SolverResultChecker {
  private SolverResultChecker() {}

  /**
   * Checks results are received from solver against expected values of variables.
   * @param solverResult The result that was received from solver.
   * @param expectedVariables The variables and their values that are expected.
   */
  public static void check(
      final SolverResult solverResult,
      final Iterable<Variable> expectedVariables) {
    checkErrors(solverResult.getErrors());

    final SolverResult.Status status = solverResult.getStatus();
    final Enum<?> satStatus = SolverResult.Status.SAT;
    Assert.assertSame("Failed to solve the constraint. Status: " + status, status, satStatus);

    final TreeMap<String, Variable> results = new TreeMap<>();
    for (Variable v : solverResult.getVariables()) {
      results.put(v.getName(), v);
    }

    final TreeMap<String, Variable> expected = new TreeMap<>();
    for (final Variable v : expectedVariables) {
      expected.put(v.getName(), v);
    }

    Assert.assertEquals("Wrong variable number", results.size(), expected.size());

    for (final Variable variable : results.values()) {

      final String varName = variable.getName();
      final Data varData = variable.getData();
      final DataType varType = varData.getType();
      final Object varValue = varData.getValue();

      final Variable expectedVar = expected.get(varName);
      final Data expData = expectedVar.getData();
      final DataType expType = expData.getType();
      final Object expValue = expData.getValue();

      Assert.assertNotNull(String.format("Unexpected variable name. '%s'", varName), expectedVar);
      Assert.assertEquals(String.format("Unexpected variable type. '%s' vs '%s'", varType, expType),
          varType, expType);
      Assert.assertEquals(
          String.format("Unexpected value of the %s variable: '%s', expected: '%s'", varName,
                  varValue, expValue),
          varValue, expValue);
    }
  }

  /**
   * Accumulates error messages and prints them in a single assertion message.
   * @param errors The error messages.
   */
  public static void checkErrors(final Iterable<String> errors) {
    final Iterator<String> errorIterator = errors.iterator();

    if (!errorIterator.hasNext()) {
      return;
    }

    final StringBuilder errorStringBuilder = new StringBuilder();
    errorStringBuilder.append("Errors occured:");

    while (errorIterator.hasNext()) {
      errorStringBuilder.append("\r\n");
      errorStringBuilder.append(errorIterator.next());
    }

    Assert.fail(errorStringBuilder.toString());
  }
}
