/*
 * Copyright 2014-2018 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.fortress.solver.constraint;

import org.junit.Assert;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;

import java.util.Iterator;

/**
 * The {@code ConstraintEqualityChecker} class contains utility methods to perform comparisons
 * of constraints and their internal parts.
 *
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public final class ConstraintEqualityChecker {
  private ConstraintEqualityChecker() {}

  /**
   * Check whether actual object is the same as the expected one.
   * @param expected The expected object.
   * @param actual The actual object.
   */
  public static void check(Constraint expected, Constraint actual) {
    Assert.assertNotNull(expected);
    Assert.assertNotNull(actual);
    Assert.assertNotSame("The same object", expected, actual);

    Assert.assertEquals("Constraint names do not match.", expected.getName(), actual.getName());
    Assert.assertSame("Constraint kinds.", expected.getKind(), actual.getKind());
    Assert.assertEquals("Constraint descriptions do not match.", expected.getDescription(),
        actual.getDescription());

    check((Formulas) expected.getInnerRep(), (Formulas) actual.getInnerRep());
  }

  /**
   * Check whether actual object is the same as the expected one.
   * @param expected The expected object.
   * @param actual The actual object.
   */
  public static void check(Formulas expected, Formulas actual) {
    Assert.assertNotNull(expected);
    Assert.assertNotNull(actual);
    Assert.assertNotSame("The same object", expected, actual);

    Assert.assertNotNull(expected.exprs());
    Assert.assertNotNull(actual.exprs());
    Assert.assertNotSame("The same object", expected.exprs(), actual.exprs());

    final Iterator<Node> expectedIterator = expected.exprs().iterator();
    final Iterator<Node> actualIterator = actual.exprs().iterator();

    while (expectedIterator.hasNext() && actualIterator.hasNext()) {
      check(expectedIterator.next(), actualIterator.next());
    }

    Assert.assertEquals("The numbers of formulas are different.", expectedIterator.hasNext(),
        actualIterator.hasNext());
  }

  /**
   * Check whether actual object is the same as the expected one.
   * @param expected The expected object.
   * @param actual The actual object.
   */
  public static void check(NodeOperation expected, NodeOperation actual) {
    Assert.assertNotNull(expected);
    Assert.assertNotNull(actual);
    Assert.assertNotSame("The same object", expected, actual);

    Assert.assertSame("Invalid element ID.", expected.getKind(), Node.Kind.OPERATION);
    Assert.assertSame("Invalid element ID.", actual.getKind(), Node.Kind.OPERATION);
    Assert.assertEquals("Different operation IDs.", expected.getOperationId(),
        actual.getOperationId());

    // TODO: Temporary requirement. Once the getDataType method is implemented to return a proper
    // value
    // this code must be replaced with a proper check.
    // Assert.assertNull(expected.getDataType());
    // Assert.assertNull(actual.getDataType());

    int operandIndex = 0;
    while (operandIndex < expected.getOperandCount()) {
      if ((null != expected.getOperand(operandIndex))
          && (null != actual.getOperand(operandIndex))) {
        check(expected.getOperand(operandIndex), actual.getOperand(operandIndex));
      }

      ++operandIndex;
    }
  }

  /**
   * Check whether actual object is the same as the expected one.
   * @param expected The expected object.
   * @param actual The actual object.
   */
  public static void check(NodeBinding expected, NodeBinding actual) {
    Assert.assertNotNull(expected);
    Assert.assertNotNull(actual);
    Assert.assertNotSame("The same object", expected, actual);

    Assert.assertSame("Invalid element ID.", expected.getKind(), Node.Kind.BINDING);
    Assert.assertSame("Invalid element ID.", actual.getKind(), Node.Kind.BINDING);
    Assert.assertEquals("Different number of bound variables.", expected.getBindings().size(),
        actual.getBindings().size());

    for (int i = 0; i < expected.getBindings().size(); ++i) {
      Assert.assertEquals("Different binding order.", getBoundName(expected, i),
          getBoundName(actual, i));
      check(getBoundValue(expected, i), getBoundValue(actual, i));
    }
  }

  /**
   * Check whether actual object is the same as the expected one.
   * @param expected The expected object.
   * @param actual The actual object.
   */
  public static void check(NodeVariable expected, NodeVariable actual) {
    Assert.assertNotNull(expected);
    Assert.assertNotNull(actual);
    Assert.assertNotSame("The same object", expected, actual);

    Assert.assertSame("Invalid node kind.", expected.getKind(), Node.Kind.VARIABLE);
    Assert.assertSame("Invalid node kind.", actual.getKind(), Node.Kind.VARIABLE);

    Assert.assertEquals("Variable names do not match.", expected.getName(), actual.getName());

    check(expected.getData().getType(), actual.getData().getType());
    if (!((null == expected.getValue()) && (null == actual.getValue()))) {
      Assert.assertEquals("Variable values do not match.", expected.getValue(), actual.getValue());
    }
  }

  /**
   * Check whether actual object is the same as the expected one.
   * @param expected The expected object.
   * @param actual The actual object.
   */
  public static void check(NodeValue expected, NodeValue actual) {
    Assert.assertNotNull(expected);
    Assert.assertNotNull(actual);
    Assert.assertNotSame("The same object", expected, actual);

    Assert.assertSame("Invalid element ID.", expected.getKind(), Node.Kind.VALUE);
    Assert.assertSame("Invalid element ID.", actual.getKind(), Node.Kind.VALUE);
    check(expected.getData().getType(), actual.getData().getType());
    Assert.assertEquals("Values do not match.", expected.getValue(), actual.getValue());
  }

  /**
   * Check whether actual object is the same as the expected one.
   * @param expected The expected object.
   * @param actual The actual object.
   */
  public static void check(Node expected, Node actual) {
    Assert.assertNotNull(expected);
    Assert.assertNotNull(actual);

    Assert.assertNotSame("The same object", expected, actual);
    Assert.assertSame("Different element IDs.", expected.getKind(), actual.getKind());

    switch (expected.getKind()) {
      case VALUE:
        check((NodeValue) expected, (NodeValue) actual);
        break;

      case VARIABLE:
        check((NodeVariable) expected, (NodeVariable) actual);
        break;

      case OPERATION:
        check((NodeOperation) expected, (NodeOperation) actual);
        break;

      case BINDING:
        check((NodeBinding) expected, (NodeBinding) actual);
        break;

      default:
        Assert.fail("Unknown element type.");
        break;
    }
  }

  /**
   * Check whether actual object is the same as the expected one.
   * @param expected The expected object.
   * @param actual The actual object.
   */
  public static void check(DataType expected, DataType actual) {
    Assert.assertNotNull(expected);
    Assert.assertNotNull(actual);

    Assert.assertSame("Data type IDs do not match.", expected.getTypeId(), actual.getTypeId());
    Assert.assertEquals("Data type sizes do not match.", expected.getSize(), actual.getSize());
  }

  private static String getBoundName(NodeBinding node, int index) {
    return node.getBindings().get(index).getVariable().getName();
  }

  private static Node getBoundValue(NodeBinding node, int index) {
    return node.getBindings().get(index).getValue();
  }
}
